/*-
 * Copyright (c) 2002, Mike Benham
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 4. Neither the name of this program nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <openssl/ssl.h>

#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/sem.h>

#include <string.h>

#include "cache.h"

static SessionCache *cache;
static int semaphore;

static void lock(int semaphore);
static void unlock(int semaphore);
static void removeSessionId(char* id, int idLen);

union semun {
  int val;                  /* value for SETVAL */
  struct semid_ds *buf;     /* buffer for IPC_STAT, IPC_SET */
  unsigned short *array;    /* array for GETALL, SETALL */
  struct seminfo *__buf;    /* buffer for IPC_INFO */
};

void CACHE_initialize() {
  union semun arg;
  struct shmid_ds ret;

  int id    = shmget(ftok("sslsniff.c", 0), sizeof(SessionCache), IPC_CREAT | 0666);
  cache     = (SessionCache*)shmat(id, 0, 0);
  semaphore = semget(ftok("sslsniff.c", 0), 1, IPC_CREAT | 0666);
  arg.val   = 1;

  if (semctl(semaphore, 0, SETVAL, arg) == -1) {
    perror("semctl");
    exit(1);
  }

  if (shmctl(id, IPC_RMID, &ret) < 0) {
    perror("shmctl");
    exit(1);
  }

  bzero(cache, sizeof(SessionCache));
}

int CACHE_setNewSessionId0(SSL *s, SSL_SESSION *session, char *id, int idLength) {
  int encodedLength = i2d_SSL_SESSION(session, NULL);

  unsigned char* b;
  int current;

  if (encodedLength > MAX_ENCODING_SIZE) {
    fprintf(stderr, "Encoded Length: %d too big for session cache, skipping.\n", encodedLength);
    return 1;
  }

  lock(semaphore);
  removeSessionId(id, idLength);

  current = cache->current;
  b       = cache->sessions[current].encoding;

  i2d_SSL_SESSION(session, &b);

  memcpy(cache->sessions[current].id, id, idLength);
  cache->sessions[current].encodingLength = encodedLength;
  cache->sessions[current].idLength       = idLength;
  cache->current                          = (current + 1) % CACHE_SIZE;

  unlock(semaphore);

  return 1;  
}

SSL_SESSION * CACHE_getSessionId(SSL *s, unsigned char *id, int idLength, int *ref) {
  int i;
  SSL_SESSION *ret;
  unsigned char *b;

  *ref = 0;

  lock(semaphore);

  for (i=0;i<CACHE_SIZE;i++) {
    if (memcmp(cache->sessions[i].id, id, idLength) == 0) {
      b = (unsigned char*)malloc(cache->sessions[i].encodingLength);
      memcpy(b, cache->sessions[i].encoding, cache->sessions[i].encodingLength);
      ret = d2i_SSL_SESSION(NULL, (const unsigned char **)&b, cache->sessions[i].encodingLength);
      unlock(semaphore);
      return ret;
    }
  }

  unlock(semaphore);

  return NULL;  
}

int CACHE_setNewSessionId(SSL *s, SSL_SESSION *session) {
  return CACHE_setNewSessionId0(s, session, session->session_id, session->session_id_length);
}

static void removeSessionId(char* id, int idLength) {
  int i;

  for (i=0;i<CACHE_SIZE;i++) {
    if (memcmp(cache->sessions[i].id, id, idLength) == 0) {
      bzero(cache->sessions[i].id, idLength);
    }
  }
}

static void lock(int semaphore) {
  struct sembuf buf = {0, -1, 0};
  if (semop(semaphore, &buf, 1) < 0) {
    fprintf(stderr, "ASSERT - semaphore aquire failed.\n");
    exit(1);
  }
}

static void unlock(int semaphore) {
  struct sembuf buf = {0, 1, 0};
  if (semop(semaphore, &buf, 1) < 0) {
    fprintf(stderr, "ASSERT - semaphore aquire failed.\n");
    exit(1);
  }
}
