/*-
 * Copyright (c) 2002, Mike Benham
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 4. Neither the name of this program nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/socket.h>
#include <sys/types.h>
#include <netdb.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "log.h"

#define POST_LENGTH 4
#define POST_STRING "POST"
#define MIN(X, Y)  ((X) < (Y) ? (X) : (Y))

static int isFinishedBuffering(Log *context);
static int isLoggableData(Log *context);
static void switchToLogging(Log *context, char *buf, int length);
static void switchToProxying(Log *context);
static void bufferData(Log *context, char* buf, int length);
static void logData(Log *context, char* buf, int length);

static void connectionString(const char *logDirectory, int client, int server, 
			     char *buf, int length);

Log * LOG_init(const char* logDirectory, int client, int server, int postOnly) {
  Log *log         = (Log*)malloc(sizeof(Log));
  log->bufferIndex = 0;
  log->status      = (postOnly ? BUFFERING : PRELOGGING);
  log->logFile     = NULL;

  connectionString(logDirectory, client, server, log->name, sizeof(log->name));

  return log;
}

void LOG_log(Log *context, char* buf, int length) {
  switch (context->status) {
  case BUFFERING:  bufferData(context, buf, length);      break;
  case LOGGING:    logData(context, buf, length);         break;
  case PRELOGGING: switchToLogging(context, buf, length); break;
  case PROXYING:                                          break;
  default:
    fprintf(stderr, "ASSERT - Unknown LOG state: %d\n", context->status);
    exit(1);
  }
}

void LOG_flush(Log *context) {
  if (context->logFile != NULL)
    fflush(context->logFile);
}

static void connectionString(const char *logDirectory, int client, int server, 
			     char *buf, int length) 
{
  char clientName[512];
  char serverName[512];

  struct sockaddr_in clientAddr;
  struct sockaddr_in serverAddr;

  unsigned int size;

  size = sizeof(clientAddr);
  if (getpeername(client, (struct sockaddr*)&clientAddr, &size) < 0) {
    perror("ASSERT - c getpeername failed.");
    exit(1);
  }

  size = sizeof(serverAddr);
  if (getpeername(server, (struct sockaddr*)&serverAddr, &size) < 0) {
    perror("ASSERT - s getpeername failed.");
    exit(1);
  }

  strcpy(clientName, inet_ntoa(clientAddr.sin_addr));
  strcpy(serverName, inet_ntoa(serverAddr.sin_addr));
  
  snprintf(buf, length, "%s/%s.%d-%s.%d", 
	   logDirectory,
	   clientName, ntohs(clientAddr.sin_port), 
	   serverName, ntohs(serverAddr.sin_port));
}

static int isFinishedBuffering(Log *context) {
  return (context->bufferIndex == POST_LENGTH);
}

static int isLoggableData(Log *context) {
  context->buffer[context->bufferIndex] = '\0';
  return (strcmp(context->buffer, POST_STRING) == 0);
}

static void switchToLogging(Log *context, char *buf, int length) {
  context->status  = LOGGING;
  context->logFile = fopen(context->name, "w");
  logData(context, buf, length);
  printf("Intercepted Connection: %s [Logging]\n", context->name);
}

static void switchToProxying(Log *context) {
  context->status = PROXYING;
  printf("Intercepted Connection: %s [Proxying]\n", context->name);
}

static void bufferData(Log *context, char* buf, int length) {
  int bytesToBuffer = MIN(POST_LENGTH - context->bufferIndex, length);
  memcpy(context->buffer, buf, bytesToBuffer);
  (context->bufferIndex) += bytesToBuffer;

  if (isFinishedBuffering(context)) {
    if (isLoggableData(context)) switchToLogging(context, buf, length);
    else                         switchToProxying(context);
  }
}

static void logData(Log *context, char* buf, int length) {
  buf[length] = '\0';
  fprintf(context->logFile, "%s", buf);
}
