/*-
 * Copyright (c) 2002, Mike Benham
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 4. Neither the name of this program nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <openssl/pem.h>
#include <openssl/conf.h>
#include <openssl/x509v3.h>
#include <openssl/ssl.h>
#include <openssl/err.h>

#include <sys/poll.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <string.h>

#include "mssl.h"
#include "cache.h"

static int forwardData(SSL *from, SSL *to, Log *logFile);
static int isAvailable(int revents);
static int isClosed(int revents);

static void getServerName(int serverFd, char* name) {
  struct sockaddr_in serverAddr;
  unsigned int size = sizeof(serverAddr);

  if (getpeername(serverFd, (struct sockaddr*)&serverAddr, &size) < 0) {
    perror("ASSERT - s getpeername failed.");
    exit(1);
  }

  strcpy(name, inet_ntoa(serverAddr.sin_addr));
}

SSL * MSSL_handshakeWithServer(int serverFd) {
  char serverName[512];
  int bogus;

  SSL_CTX *serverCtx;
  SSL *serverSession;
  SSL_SESSION *sessionId;

  getServerName(serverFd, serverName);

  serverCtx     = SSL_CTX_new(SSLv23_client_method());
  serverSession = SSL_new(serverCtx);
  sessionId     = CACHE_getSessionId(serverSession, serverName, strlen(serverName), &bogus);

  if (sessionId != NULL) {
    SSL_set_session(serverSession, sessionId);
    SSL_SESSION_free(sessionId);
  }

  SSL_set_connect_state(serverSession);
  SSL_set_fd(serverSession, serverFd);
  SSL_set_options(serverSession, SSL_OP_ALL);
  
  if (SSL_connect(serverSession) < 0) {
    fprintf(stderr, "Error on SSL Connect.\n");
    exit(1);
  }

  CACHE_setNewSessionId0(serverSession, SSL_get1_session(serverSession), serverName, strlen(serverName));

  return serverSession;
}

SSL * MSSL_handshakeWithClient(int client, X509 *spoofedCert, Credentials *credentials) {
  SSL_CTX *clientContext = SSL_CTX_new(SSLv23_server_method());
  SSL *clientSession;

  SSL_CTX_sess_set_new_cb(clientContext, CACHE_setNewSessionId);
  SSL_CTX_sess_set_get_cb(clientContext, CACHE_getSessionId);

  SSL_CTX_use_certificate(clientContext, spoofedCert);
  SSL_CTX_use_PrivateKey(clientContext, credentials->leafKey);

  if (SSL_CTX_check_private_key(clientContext) == 0) {
    fprintf(stderr, "*** Assertion Failed - Generated PrivateKey Doesn't Work.\n");
    exit(1);
  }

  SSL_CTX_add_extra_chain_cert(clientContext, credentials->middleCertificate);
  SSL_CTX_set_mode(clientContext, SSL_MODE_AUTO_RETRY);

  clientSession = SSL_new(clientContext);
  SSL_set_fd(clientSession, client);

  if (SSL_accept(clientSession) == 0) {
    fprintf(stderr, "SSL Accept Failed!\n");
    exit(1);
  }

  return clientSession;
}


void MSSL_shuttleData(int client, SSL *clientSession, int server, SSL  *serverSession, Log *log) {
  struct pollfd fds[2] = {{client, POLLIN | POLLPRI | POLLHUP | POLLERR, 0},
			  {server, POLLIN | POLLPRI | POLLHUP | POLLERR, 0}};

  for (;;) {
    if (poll(fds, 2, -1) < 0) return;

    if (isAvailable(fds[0].revents)) if (forwardData(clientSession, serverSession, log) != 0) return;
    if (isAvailable(fds[1].revents)) if (forwardData(serverSession, clientSession, log) != 0) return;
    if (isClosed(fds[0].revents)) return;
    if (isClosed(fds[0].revents)) return;
  }
}

static int isAvailable(int revents) {
  return revents & POLLIN || revents & POLLPRI;
}

static int isClosed(int revents) {
  return revents & POLLERR || revents & POLLHUP;
}

static int forwardData(SSL *from, SSL *to, Log *log) {
  char buf[4096];
  int bytesRead; 
  int bytesWritten;

  do {
    if ((bytesRead    = SSL_read(from, buf, sizeof(buf))) <= 0)     return -1;
    if ((bytesWritten = SSL_write(to, buf, bytesRead)) < bytesRead) return -1;

    LOG_log(log, buf, bytesRead);
  } while (SSL_pending(from));

  LOG_flush(log);

  return 0;
}
